from collections.abc import Mapping


class Record(Mapping):

    __slots__ = ('_names', '_values')

    def __init__(self, names, values):
        self._names = names
        self._values = values

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    def __getitem__(self, item):
        if isinstance(item, str):
            try:
                item = self._names.index(item)
            except ValueError:
                raise KeyError(item)
        return self._values[item]

    def get(self, item, default=None):
        try:
            return self[item]
        except (KeyError, IndexError):
            return default

    def keys(self):
        return self._names

    def values(self):
        return self._values

    def items(self):
        return zip(self._names, self._values)

    def __eq__(self, other):
        if isinstance(other, Mapping):
            return super().__eq__(other)
        else:
            return self._values == list(other)

    def __repr__(self):
        return '<Record {}>'.format(
            ' '.join(
                f'{name}={value!r}'
                for name, value in self.items()
            )
        )



import logging
import simplejson as json
from typing import Any, AsyncGenerator, Iterable, List, Optional

import aiohttp

from .compiler import Compiler, Statement
from .dialect import ClickhouseSaDialect
from .exc import DBException, ProtocolError, exc_message_re
from .parser import parse_json_compact, JSONDecodeError
from .record import Record
from .types import TypeRegistry


logger = logging.getLogger(__name__)
sql_logger = logging.getLogger(f'{__name__}.SQL')


class Client:

    def __init__(
        self, session: aiohttp.ClientSession, *, url='http://localhost:8123/',
        user=None, password=None, database='default', compress_response=False,
        dialect=None, types=None, **settings,
    ):
        self._session = session
        self.url = url
        self.params = {}
        if user:
            self.params["user"] = user
        if password:
            self.params["password"] = password
        if database:
            self.params["database"] = database
        if compress_response:
            self.params["enable_http_compression"] = 1
        self.params.update(settings)
        if dialect is None: # pragma: no cover
            # XXX Do we actualy need the ability to pass custom dialect?
            dialect = ClickhouseSaDialect()
        if types is None:
            types = TypeRegistry()
        self._types = types
        self._compiler = Compiler(dialect=dialect, escape=types.escape)

    async def _execute(self, statement: Statement, *args) -> Iterable[Record]:
        compiled, json_each_row_parameters = self._compiler.compile_statement(
            statement, args,
        )
        sql_logger.debug(compiled)
        compiled_with_params = compiled
        rows = None
        if json_each_row_parameters:
            to_json = self._types.to_json # lookup optimization
            rows = [
                json.dumps(
                    {name: to_json(value) for name, value in row.items()},
                    use_decimal=True,
                )
                for row in json_each_row_parameters
            ]
            if sql_logger.isEnabledFor(logging.DEBUG):
                for idx, row in enumerate(rows):
                    sql_logger.debug(f'{idx}: {row}')
            compiled_with_params += '\n' + '\n'.join(rows)

        # First attempt may fail due to broken state of aiohttp session
        # (aiohttp doesn't handle connection closing properly?)
        for retrying in [False, True]:
            try:
                async with self._session.post(
                    self.url,
                    params = {'default_format': 'JSONCompact', **self.params},
                    data = compiled_with_params.encode(),
                ) as response:
                    body = await response.read()
                    if response.status != 200:
                        raise DBException.from_message(
                            body.decode(errors='replace'),
                            statement=compiled, rows=rows,
                        )

                    elif response.content_type == 'application/json':
                        try:
                            return parse_json_compact(self._types, body)
                        except JSONDecodeError:
                            body_str = body.decode(errors='replace')
                            m = exc_message_re.search(body_str)
                            if not m:
                                raise
                            raise DBException.from_message(
                                body_str[m.start():],
                                statement=compiled, rows=rows,
                            )
                    else:
                        return ()
            except aiohttp.ClientError as exc:
                if retrying:
                    raise ProtocolError(exc) from exc
                logger.debug(f'First attempt failed, retrying (error: {exc})')

        assert False, 'Unreachable'  # To silence mypy

    async def iterate(
        self, statement: Statement, *args,
    ) -> AsyncGenerator[Record, None]:
        for row in await self._execute(statement, *args):
            yield row

    async def execute(self, statement: Statement, *args) -> None:
        await self._execute(statement, *args)

    async def fetch(self, statement: Statement, *args) -> List[Record]:
        return list(await self._execute(statement, *args))

    async def fetchrow(self, statement: Statement, *args) -> Optional[Record]:
        gen = await self._execute(statement, *args)
        return next(iter(gen), None)

    async def fetchval(self, statement: Statement, *args) -> Any:
        row = await self.fetchrow(statement, *args)
        if row is not None:
            return row[0]

    def __await__(self):
        # For compartibility with asyncpg (`await pool.acquire(...)`)
        yield from []
        return self

    async def close(self):
        # For compartibility with asyncpg
        pass

    # Allow using client as context manager when returned from `Pool.acquire()`
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



from pkg_resources import get_distribution, DistributionNotFound

from .client import Client
from .exc import DBException, ProtocolError
from .pool import connect, create_pool
from .sql import select

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass



from collections import namedtuple
import re


# Based on `getExceptionMessage()`:
# https://github.com/ClickHouse/ClickHouse/blob/master/src/Common/Exception.cpp#L370-L373
exc_message_re = re.compile(
    r'Code: (?P<code>\d+), '
        r'e\.displayText\(\) = (?P<display_text>.+?)'
        r'(?:, Stack trace[^:]*:\s+(?P<stack_trace>.+))?$',
    re.M,
)

at_row_re = re.compile(r'[(]at row (?P<num>\d+)[)]')

class AiochsaException(Exception):
    """ Base class for aiochsa exceptions """


class ProtocolError(AiochsaException):
    """ Error communicating to Clickhouse server """


RowInfo = namedtuple('RowInfo', ['num', 'content'])


class DBException(AiochsaException):
    """ Error returned from Clickhouse database """

    def __init__(
        self, code, display_text, stack_trace=None, statement=None, row=None,
    ):
        super().__init__(code, display_text, stack_trace)
        self.code = code
        self.display_text = display_text
        self.stack_trace = stack_trace
        self.statement = statement
        self.row = row

    def __str__(self):
        message = f'[Code={self.code}] {self.display_text}'
        if self.statement:
            statement = self.statement
            if len(statement) > 200:
                statement = statement[:200] + '...'
            message += f'\n{statement}'
        if self.row:
            message += f'\n{self.row.num}: {self.row.content}'
        return message

    @classmethod
    def from_message(cls, exc_message, *, statement=None, rows=None):
        m = exc_message_re.match(exc_message)
        if m:
            display_text = m.group('display_text')

            row = None
            if rows:
                at_row_m = at_row_re.search(display_text)
                if at_row_m:
                    # It's 1-based
                    row_num = int(at_row_m.group('num'))
                    if len(rows) >= row_num:
                        row = RowInfo(row_num, rows[row_num - 1])

            return cls(
                code=int(m.group('code')),
                display_text=display_text,
                stack_trace=m.group('stack_trace'),
                statement=statement,
                row=row,
            )
        else: # pragma: nocover
            return cls(
                code=None, display_text=exc_message, statement=statement,
            )



import asyncio
import pytest

from aiochsa import DBException, error_codes
from aiochsa.pool import dsn_to_params, create_pool



def test_dsn_all_explicit():
    params = dsn_to_params('clickhouse://bob:secret@pet:12345/warehouse')
    assert params == {
        'url': 'http://pet:12345',
        'database': 'warehouse',
        'user': 'bob',
        'password': 'secret',
    }


def test_dsn_wrong_scheme():
    with pytest.raises(ValueError):
        dsn_to_params('postgresql://db')


def test_dsn_default_port():
    params = dsn_to_params('clickhouse://pet')
    assert params['url'] == 'http://pet:8123'


async def test_create_pool_close(dsn):
    pool = await create_pool(dsn)
    await pool.execute('SELECT 1')
    await pool.close()


async def test_pool_async_with(pool):
    async with pool.acquire() as conn:
        await conn.execute('SELECT 1')


async def test_pool_acquire_release(pool):
    conn = await pool.acquire()
    await conn.execute('SELECT 1')
    await pool.release(conn)


LONG_QUERY = 'SELECT sleep(1) FROM numbers(10) SETTINGS max_block_size=1'


async def test_pool_params(dsn):
    async with create_pool(dsn, max_execution_time=1) as conn:
        with pytest.raises(DBException) as exc_info:
            await conn.execute(LONG_QUERY)
    assert exc_info.value.code == error_codes.TIMEOUT_EXCEEDED


async def test_pool_dsn_params(dsn):
    dsn += '?max_execution_time=1'
    async with create_pool(dsn) as conn:
        with pytest.raises(DBException) as exc_info:
            await conn.execute(LONG_QUERY)
    assert exc_info.value.code == error_codes.TIMEOUT_EXCEEDED


async def test_session_timeout(dsn):
    async with create_pool(dsn, session_timeout=0.1) as conn:
        with pytest.raises(asyncio.TimeoutError):
            await conn.execute(LONG_QUERY)



from datetime import date, datetime
from decimal import Decimal
from enum import Enum

import pytest
import sqlalchemy as sa

import aiochsa

async def test_ddl(conn, table_test):
    await conn.execute(sa.DDL(f'DROP TABLE {table_test.name}'))


@pytest.mark.xfail(
    raises=TypeError,
    # Is the game worth the candle? For it compilation can't be separated from
    # execution.
    reason='Execution of default is not supported yet',
)
async def test_execute_default(conn, table_test):
    ts_before = datetime.utcnow().replace(microsecond=0)
    now = await conn.fetchval(table_test.c.timestamp.default)
    ts_after = datetime.utcnow().replace(microsecond=0)
    assert ts_before <= now <= ts_after


async def test_func(conn):
    ts_before = datetime.utcnow().replace(microsecond=0)
    now = await conn.fetchval(sa.func.now())
    ts_after = datetime.utcnow().replace(microsecond=0)
    assert ts_before <= now <= ts_after


async def test_func_params(conn):
    result = await conn.fetchval(
        sa.func.plus(sa.bindparam('a'), sa.bindparam('b'))
            .params({'a': 12, 'b': 23})
    )
    assert result == 35


async def test_func_params_args(conn):
    result = await conn.fetchval(
        sa.func.plus(sa.bindparam('a'), sa.bindparam('b')),
        {'a': 12, 'b': 23},
    )
    assert result == 35


async def test_simple_round(conn, table_test):
    now = datetime.utcnow().replace(microsecond=0)
    values = {
        'id': 1,
        'enum': 'ONE',
        'name': 'test',
        'timestamp': now,
        'amount': Decimal('1.23'),
    }
    await conn.execute(
        table_test.insert()
            .values(values)
    )

    rows = await conn.fetch(table_test.select())
    row1 = await conn.fetchrow(table_test.select())
    row2 = await conn.fetchrow(f'SELECT * from {table_test.name}')
    assert rows == [row1]
    assert row1 == row2 == values


async def test_non_ascii(conn, table_test, any_select):
    sample = 'зразок'
    await conn.execute(
        table_test.insert()
            .values(id=1, name=sample)
    )
    name = await conn.fetchval(
        any_select([table_test.c.name])
            .where(table_test.c.id == 1)
    )
    assert name == sample


async def test_enum(conn, table_test, any_select):

    class EnumType(str, Enum):
        ONE = 'ONE'
        TWO = 'TWO'

    await conn.execute(
        table_test.insert()
            .values(id=1, enum=EnumType.ONE)
    )
    value = await conn.fetchval(
        any_select([table_test.c.enum])
            .where(table_test.c.id == 1)
    )
    assert value == EnumType.ONE


async def test_unsupported_type(conn, any_select):
    with pytest.raises(TypeError):
        await conn.fetchval(
            any_select([sa.bindparam('a')])
                .params(a=...)
        )


@pytest.mark.xfail(reason='Feature is disabled in favour of speed')
async def test_defaults(conn, table_test):
    ts_before = datetime.utcnow().replace(microsecond=0)
    values = {
        'id': 1,
        'enum': 'ONE',
        'name': 'test',
    }
    await conn.execute(
        table_test.insert()
            .values(values)
    )
    ts_after = datetime.utcnow().replace(microsecond=0)

    row = await conn.fetchrow(table_test.select())
    for field, value in values.items():
        assert row[field] == value
    assert ts_before <= row['timestamp'] <= ts_after
    assert row['amount'] == Decimal(0)


async def test_insert_multiple(conn, table_test, any_select):
    values = [
        {
            'id': i + 1,
            'name': f'test{i + 1}',
            'enum': 'TWO' if i % 2 else 'ONE',
        }
        for i in range(6)
    ]
    await conn.execute(
        table_test.insert()
            .values(values)
    )

    rows = await conn.fetch(
        any_select([table_test.c.id])
            .where(table_test.c.enum == 'ONE')
            .order_by(table_test.c.id.desc())
    )
    assert [item_id for (item_id,) in rows] == [5, 3, 1]


async def test_insert_multiple_args(conn, table_test, any_select):
    values = [
        {'id': i + 1, 'name': f'test{i + 1}'}
        for i in range(3)
    ]
    await conn.execute(
        table_test.insert(), *values,
    )

    rows = await conn.fetch(
        any_select([table_test.c.id, table_test.c.name])
    )
    assert rows == values


async def test_insert_select(conn, table_test, table_mt, any_select):
    values = [
        {'id': i + 1, 'name': f'test{i + 1}'}
        for i in range(3)
    ]
    await conn.execute(
        table_test.insert(), *values,
    )

    await conn.execute(
        table_mt.insert()
            .from_select(
                [table_mt.c.num, table_mt.c.title],
                any_select([table_test.c.id, table_test.c.name])
                    .where(table_test.c.id > 1)
            )
    )
    rows = await conn.fetch(
        table_mt.select()
    )
    assert rows == [(2, 'test2'), (3, 'test3')]


@pytest.mark.xfail(reason='CTE is not supported by clickhouse-sqlalchemy')
async def test_insert_select_cte(conn, table_mt, any_select):
    await conn.execute(
        table_mt.insert(), {'title': 'test', 'num': 1},
    )

    max_num = (
        any_select([sa.func.max(table_mt.c.num)])
            .cte()
    )

    await conn.execute(
        table_mt.insert()
            .from_select(
                [table_mt.c.num, table_mt.c.title],
                any_select([max_num, table_mt.c.title])
            )
    )
    rows = await conn.fetch(
        table_mt.select()
    )
    assert rows == [('test', 1), ('test', 2)]


async def test_join(conn, table_test, any_select):
    await conn.execute(
        table_test.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    test_alias = table_test.alias()
    rows = await conn.fetch (
        any_select([table_test.c.id, test_alias.c.id])
            .select_from(
                table_test.join(
                    test_alias,
                    test_alias.c.id == table_test.c.id,
                )
            )
    )
    assert {tuple(row) for row in rows} == {(1, 1), (2, 2), (3, 3)}


async def test_iterate(conn, table_test, any_select):
    await conn.execute(
        table_test.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    rows_agen = conn.iterate(
        any_select([table_test.c.id])
    )
    assert [item_id async for (item_id,) in rows_agen] == [1, 2, 3]


async def test_fetchval_empty(conn, table_test, any_select):
    value = await conn.fetchval(
        any_select([table_test.c.id])
    )
    assert value is None


async def test_select_params(conn, table_test, any_select):
    await conn.execute(
        table_test.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    rows = await conn.fetch(
        any_select([table_test.c.id])
            .where(table_test.c.id >= sa.bindparam('min_id'))
            .params(min_id=2)
    )
    assert [item_id for (item_id,) in rows] == [2, 3]


async def test_select_params_args(conn, table_test, any_select):
    await conn.execute(
        table_test.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    rows = await conn.fetch(
        any_select([table_test.c.id])
            .where(table_test.c.id >= sa.bindparam('min_id')),
        {'min_id': 2},
    )
    assert [item_id for (item_id,) in rows] == [2, 3]


async def test_final_hint(conn, table_smt):
    await conn.execute(
        table_smt.insert(),
        *[{'key': 123, 'value': i} for i in range(10)]
    )
    rows = await conn.fetch(
        table_smt.select()
            .with_hint(table_smt, 'FINAL')
    )
    assert len(rows) == 1
    assert rows[0]['value'] == 45


@pytest.mark.parametrize('alised', [False, True])
async def test_join_final_hint(
    conn, table_test, table_smt, alised, any_select,
):
    await conn.execute(
        table_test.insert(),
        {'id': 12, 'name': 'name12'},
        {'id': 23, 'name': 'name23'},
    )
    await conn.execute(
        table_smt.insert(),
        *[{'key': 12, 'value': i} for i in range(10)]
    )

    if alised:
        table_smt = table_smt.alias()
    query = (
        any_select([table_test.c.name, table_smt.c.value])
            .select_from(
                table_smt.join(
                    table_test,
                    table_test.c.id == sa.func.toUInt64(table_smt.c.key),
                )
            )
            .with_hint(table_smt, 'FINAL')
    )
    rows = await conn.fetch(query)
    assert len(rows) == 1
    assert rows[0]['name'] == 'name12'


async def test_aggregate_function(conn, table_amt, any_select):
    await conn.execute('INSERT INTO test_amt SELECT 1, sumState(123)')
    value = await conn.fetchval(
        any_select([table_amt.c.value])
    )
    assert isinstance(value, aiochsa.types.AggregateFunction)
    value = await conn.fetchval(
        any_select([sa.func.sumMerge(table_amt.c.value)])
    )
    assert value == 123


async def test_nested_structures(conn):
    value = await conn.fetchval(
        r"SELECT ("
            r"1, "
            r"(2, NULL, 'a\nb\tc\0d\'', ['a', '\'']),"
            r"1.23, toDecimal32('1.23', 2), toDate('2000-01-01')"
        r")"
    )
    assert value == (
        1,
        (2, None, 'a\nb\tc\0d\'', ['a', '\'']),
        1.23, Decimal('1.23'), date(2000, 1, 1)
    )


async def test_limit_by_with_order_by(conn, table_mt):
    await conn.execute(
        table_mt.insert(),
        [
            {'num': 1, 'title': '13'},
            {'num': 1, 'title': '12'},
            {'num': 1, 'title': '11'},
            {'num': 2, 'title': '21'},
            {'num': 2, 'title': '22'},
        ],
    )

    query = (
        aiochsa.select([table_mt.c.title])
            .order_by(table_mt.c.num, table_mt.c.title)
    )
    rows = await conn.fetch(
        query.limit_by([table_mt.c.num], limit=2)
    )
    assert [title for (title,) in rows] == ['11', '12', '21', '22']

    rows = await conn.fetch(
        query.limit_by([table_mt.c.num], offset=1, limit=1)
    )
    assert [title for (title,) in rows] == ['12', '22']


async def test_limit_by_without_order_by(conn, table_smt):
    await conn.execute(
        table_smt.insert(),
        [
            {'key': 1, 'value': 13},
            {'key': 1, 'value': 12},
            {'key': 1, 'value': 11},
            {'key': 2, 'value': 21},
            {'key': 2, 'value': 22},
        ],
    )
    rows = await conn.fetch(
        aiochsa.select([table_smt.c.key])
            .limit_by([table_smt.c.key], limit=1)
    )
    assert {value for (value,) in rows} == {1, 2}



from types import SimpleNamespace
from typing import Union

from sqlalchemy.engine.util import _distill_params
from sqlalchemy.sql import func, ClauseElement
from sqlalchemy.sql.ddl import DDLElement
from sqlalchemy.sql.dml import Insert
from sqlalchemy.sql.functions import FunctionElement


Statement = Union[str, ClauseElement]


class Compiler:

    def __init__(self, dialect, escape):
        self._dialect = dialect
        self._escape = escape

    def _execute_clauseelement(self, elem, multiparams):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_clauseelement`
        # (event signaling, caching are removed; separate parameters are merge into
        # clause element)

        if not multiparams and isinstance(elem, Insert) and elem.parameters:
            # We disable normal way of extracting parameters, so we have to
            # handle parameters incorporated into clause with
            # `Insert.values(...)` call manually.
            if isinstance(elem.parameters, dict):
                multiparams = (elem.parameters,)
            else:
                multiparams = tuple(elem.parameters)

        distilled_params = _distill_params(multiparams, {})

        init_compiled_parameters = distilled_params
        if isinstance(elem, Insert):
            # `_init_compiled()` method creates placeholders for each row in
            # `multiparams`.  This is very slow and we don't need them when
            # passing separately in JSONEachRow format.  As workaroung we
            # pretend that we don't have parameters at all.
            init_compiled_parameters = ()

        compiled_sql = elem.compile(
            dialect=self._dialect,
            inline=True, # Never add constructs to return default values
        )
        return self._execute_context(
            self._dialect,
            self._dialect.execution_ctx_cls._init_compiled,
            compiled_sql,
            distilled_params,
            # The rest are `*args` to `_init_compiled`
            compiled_sql,
            init_compiled_parameters,
        )

    def _execute_function(self, func, multiparams):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_function`
        return self._execute_clauseelement(func.select(), multiparams)

    def _execute_ddl(self, ddl, multiparams):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_ddl` (event
        # signaling is removed).
        compiled = ddl.compile(dialect=self._dialect)
        return self._execute_context(
            self._dialect,
            self._dialect.execution_ctx_cls._init_ddl,
            compiled,
            None,
            compiled,
        )

    def _execute_context(self, dialect, constructor, statement, parameters, *args):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_context` (event
        # signaling is removed; return statement as string instead of dispatching
        # to `dialect.{do_execute,do_executemany,do_execute_no_params}`; no
        # exception convertion).
        conn = SimpleNamespace(dialect=dialect, _execution_options={})
        db_api_conn = SimpleNamespace(cursor=lambda: None)
        context = constructor(dialect, conn, db_api_conn, *args)

        # Only SQL compiler has this attribute, but not DDL compiler
        if getattr(statement, '_clickhouse_json_each_row', False):
            # We can't use `context.parameters` here, since we trick
            # `_init_compiled()` to think we have no parameters, meaning
            # they're always empty here.
            return context.statement, parameters or context.parameters
        else:
            assert len(context.parameters) == 1
            escaped = {
                name: self._escape(value)
                for name, value in context.parameters[0].items()
            }
            return context.statement % escaped, ()


    def compile_statement(self, statement: Statement, args):
        if isinstance(statement, str):
            assert not args
            return statement, args
        elif isinstance(statement, ClauseElement):
            if isinstance(statement, DDLElement):
                statement, parameters = self._execute_ddl(statement, args)
            elif isinstance(statement, FunctionElement):
                statement, parameters = self._execute_function(statement, args)
            else:
                statement, parameters = self._execute_clauseelement(
                    statement, args,
                )
            return statement, parameters
        else:
            raise TypeError(f'Execution of {type(statement)} is not supported')
